(*
  Ita IDE Plugin

  Copyright (c) 2014-2015 Lyna

  This software is provided 'as-is', without any express or implied
  warranty. In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, subject to the following restrictions:

    1. The origin of this software must not be misrepresented; you must not
    claim that you wrote the original software. If you use this software
    in a product, an acknowledgment in the product documentation would be
    appreciated but is not required.

    2. Altered source versions must be plainly marked as such, and must not be
    misrepresented as being the original software.

    3. This notice may not be removed or altered from any source
    distribution.
*)
unit UItaMain;

interface

uses
  Winapi.Windows, System.SysUtils, System.Classes, System.Generics.Collections,
  System.IniFiles, System.Win.Registry, Vcl.Controls, Vcl.Graphics, Vcl.Menus,
  Vcl.Imaging.jpeg, Vcl.Imaging.pngimage, ToolsAPI, UItaConfig;

procedure LoadBackgroundImage;
procedure SaveSettings;

var
  ImageFileName: string;
  ImagePos: TPoint = (X: 600; Y: 0);
  ImageAlpha: Byte;

implementation

const
  sEVFillRect = '@Editorcontrol@TCustomEditControl@EVFillRect$qqrrx18System@Types@TRect';
  sEVScrollRect = '@Editorcontrol@TCustomEditControl@EVScrollRect$qqrp18System@Types@TRectt1ii';
  sEditControlList = '@Editorcontrol@EditControlList';

  EVFillRectCodes: array[0..5] of Byte = (
    $53,      // PUSH EBX
    $56,      // PUSH ESI
    $8B, $F2, // MOV ESI,EDX
    $8B, $D8  // MOV EBX,EAX
  );
  EVScrollRectCodes: array[0..5] of Byte = (
    $55,          // PUSH EBP
    $8B, $EC,     // MOV EBP,ESP
    $83, $C4, $FF // ADD ESP,-xx
  );

  CoreIdeModuleName = {$IFDEF VER230}'coreide160.bpl'{$ENDIF} // XE2
                      {$IFDEF VER240}'coreide170.bpl'{$ENDIF} // XE3
                      {$IFDEF VER250}'coreide180.bpl'{$ENDIF} // XE4
                      {$IFDEF VER260}'coreide190.bpl'{$ENDIF} // XE5
                      {$IFDEF VER270}'coreide200.bpl'{$ENDIF} // XE6
                      {$IFDEF VER280}'coreide210.bpl'{$ENDIF} // XE7
                      {$IFDEF VER290}'coreide220.bpl'{$ENDIF} // XE8
                      ;

  HighlightRegKey =
    {$IFDEF VER230}'\Software\Embarcadero\BDS\9.0\Editor\Highlight\'{$ENDIF}  // XE2
    {$IFDEF VER240}'\Software\Embarcadero\BDS\10.0\Editor\Highlight\'{$ENDIF} // XE3
    {$IFDEF VER250}'\Software\Embarcadero\BDS\11.0\Editor\Highlight\'{$ENDIF} // XE4
    {$IFDEF VER260}'\Software\Embarcadero\BDS\12.0\Editor\Highlight\'{$ENDIF} // XE5
    {$IFDEF VER270}'\Software\Embarcadero\BDS\14.0\Editor\Highlight\'{$ENDIF} // XE6
    {$IFDEF VER280}'\Software\Embarcadero\BDS\15.0\Editor\Highlight\'{$ENDIF} // XE7
    {$IFDEF VER290}'\Software\Embarcadero\BDS\16.0\Editor\Highlight\'{$ENDIF} // XE8
    ;
  HighlightRegName = 'Background Color New';

var
  TCustomEditControl_EVScrollRect: Pointer;
  TCustomEditControl_EVFillRect: Pointer;
  EditControlList: ^TList;

  ImageSize: TSize;
  BufferDict: TObjectDictionary<TColor,TBitmap>;
  TempBuffer: TBitmap;

type
  TEditControl = class(TCustomControl);
procedure HookEVFillRect(Self: TEditControl; const R: TRect);
var
  rt: TRect;
  buf: TBitmap;
begin
  rt := Bounds(ImagePos.X, ImagePos.Y, ImageSize.cx, ImageSize.cy);

  if not IntersectRect(rt, rt, R) then
  begin
    Self.Canvas.FillRect(R);
    Exit;
  end;

  if rt = R then
  begin
    if BufferDict.TryGetValue(Self.Canvas.Brush.Color, buf) then
      Self.Canvas.CopyRect(rt, buf.Canvas, Bounds(rt.Left-ImagePos.X, rt.Top-ImagePos.Y, rt.Right-rt.Left, rt.Bottom-rt.Top))
    else
      FillRect(Self.Canvas.Handle, R, Self.Canvas.Brush.Handle);
  end
  else begin
    if (TempBuffer.Width < R.Width) or (TempBuffer.Height < R.Height) then
      TempBuffer.SetSize(R.Width, R.Height);

    TempBuffer.Canvas.Brush.Color := Self.Canvas.Brush.Color;
    TempBuffer.Canvas.FillRect(Rect(0, 0, R.Width, R.Height));

    if BufferDict.TryGetValue(Self.Canvas.Brush.Color, buf) then
    begin
      TempBuffer.Canvas.CopyRect(Bounds(rt.Left-R.Left, rt.Top-R.Top, rt.Width, rt.Height),
        buf.Canvas, Bounds(rt.Left-ImagePos.X, rt.Top-ImagePos.Y, rt.Width, rt.Height));
    end;

    Self.Canvas.CopyRect(R, TempBuffer.Canvas, Bounds(0, 0, R.Right-R.Left, R.Bottom-R.Top));
  end;
end;

procedure HookEVScrollRect(Self: TWinControl; const Src, Dest: TRect; DX, DY: Integer);
begin
  Self.Invalidate;
end;

procedure RefreshEditors;
var
  iEditor: Integer;
begin
  if EditControlList = nil then Exit;

  for iEditor := 0 to EditControlList^.Count-1 do
    TWinControl(EditControlList^[iEditor]).Invalidate;
end;

function Hook: Boolean;

  function IsValidCodes(P: PByte; const Codes: array of Byte): Boolean;
  var
    i: Integer;
  begin
    if P = nil then Exit(False);
    for i := Low(Codes) to High(Codes) do
    begin
      if P^ and Codes[i] <> P^ then
        Exit(False);
      Inc(p);
    end;
    Result := True;
  end;

  procedure JmpHookProc(Target: Pointer; ReplaceProc: Pointer);
  var
    oldProtect: Cardinal;
  begin
    VirtualProtect(Target, 5, PAGE_READWRITE, oldProtect);
    PByte(Target)^ := $E9;
    PInteger(Integer(Target)+1)^ := Integer(ReplaceProc) - Integer(Target) - 5;
    VirtualProtect(Target, 5, oldProtect, oldProtect);
    FlushInstructionCache(GetCurrentProcess, Target, 5);
  end;

var
  hModule: THandle;
begin
  Result := False;

  hModule := GetModuleHandle(CoreIdeModuleName);
  if hModule = 0 then Exit;

  TCustomEditControl_EVFillRect := GetProcAddress(hModule, sEVFillRect);
  TCustomEditControl_EVScrollRect := GetProcAddress(hModule, sEVScrollRect);
  EditControlList := GetProcAddress(hModule, sEditControlList);

  if not IsValidCodes(TCustomEditControl_EVFillRect, EVFillRectCodes) or
     not IsValidCodes(TCustomEditControl_EVScrollRect, EVScrollRectCodes) then
  begin
    TCustomEditControl_EVFillRect := nil;
    TCustomEditControl_EVScrollRect := nil;
    EditControlList := nil;
    Exit;
  end;

  JmpHookProc(TCustomEditControl_EVFillRect, @HookEVFillRect);
  JmpHookProc(TCustomEditControl_EVScrollRect, @HookEVScrollRect);

  RefreshEditors;

  Result := True;
end;
                                                                                  
procedure Unhook;

  procedure UnhookProc(Target: Pointer; Size: Integer; const OrgCodes: array of Byte);
  var
    oldProtect: Cardinal;
  begin
    VirtualProtect(Target, Size, PAGE_READWRITE, oldProtect);
    Move(OrgCodes[0], Target^, Size);
    VirtualProtect(Target, Size, oldProtect, oldProtect);
    FlushInstructionCache(GetCurrentProcess, Target, 5);
  end;

begin
  if TCustomEditControl_EVFillRect = nil then Exit;
  if TCustomEditControl_EVScrollRect = nil then Exit;

  UnhookProc(TCustomEditControl_EVFillRect, 5, EVFillRectCodes);
  UnhookProc(TCustomEditControl_EVScrollRect, 5, EVScrollRectCodes);

  RefreshEditors;
end;

procedure LoadSettings;
var
  ini: TMemIniFile;
begin
  ini := TMemIniFile.Create(ExtractFilePath(GetModuleName(HInstance)) + 'itaide.ini');
  try
    ImageFileName := ini.ReadString('Image', 'FileName', ImageFileName);
    ImagePos.X := ini.ReadInteger('Image', 'Left', ImagePos.X);
    ImagePos.Y := ini.ReadInteger('Image', 'Top', ImagePos.Y);
    ImageAlpha := ini.ReadInteger('Image', 'Alpha', High(Byte));
  finally
    ini.Free;
  end;
end;

procedure SaveSettings;
var
  ini: TMemIniFile;
begin
  ini := TMemIniFile.Create(ExtractFilePath(GetModuleName(HInstance)) + 'itaide.ini');
  try
    ini.WriteString('Image', 'FileName', ImageFileName);
    ini.WriteInteger('Image', 'Left', ImagePos.X);
    ini.WriteInteger('Image', 'Top', ImagePos.Y);
    ini.WriteInteger('Image', 'Alpha', ImageAlpha);
    ini.UpdateFile;
  finally
    ini.Free;
  end;
end;

procedure LoadBackgroundImage;

  procedure LoadImage(Image: TPngImage);
  var
    ext: string;
    png: TPngImage;
    jpg: TJPEGImage;
    bmp: TBitmap;
    i, x, y: Integer;
    TransparencyChunk: TChunktRNS;
    alpha: Double;
  begin
    ext := LowerCase(ExtractFileExt(ImageFileName));
    if (not FileExists(ImageFileName)) or (ext = '.png') then
    begin
      png := TPNGImage.Create;
      try
        if not FileExists(ImageFileName) then
          png.LoadFromResourceName(HInstance, 'PNG_DEFAULT')
        else
          png.LoadFromFile(ImageFileName);
        if png.Header.ColorType = COLOR_RGBALPHA then
        begin
          Image.Assign(png);
          if ImageAlpha <> 255 then
          begin
            alpha := ImageAlpha / 255;
            for y := 0 to Image.Height-1 do
              for x := 0 to Image.Width-1 do
                Image.AlphaScanline[y]^[x] := Trunc(Image.AlphaScanline[y]^[x] * alpha);
          end;
        end
        else begin
          Image.Assign(png);
          case Image.Header.ColorType of
            COLOR_GRAYSCALE, COLOR_RGB:
            begin
              Image.CreateAlpha;
              for i := 0 to Image.Height-1 do
                FillChar(Image.AlphaScanline[i]^, Image.Width, ImageAlpha);
            end;
            COLOR_PALETTE:
            begin
              Image.CreateAlpha;
              TransparencyChunk := TChunktRNS(Image.Chunks.ItemFromClass(TChunktRNS));
              FillChar(TransparencyChunk.PaletteValues[0], 256, ImageAlpha);
            end;
          end;
        end;
      finally
        png.Free;
      end;
    end
    else if (ext = '.jpg') or (ext = '.jpeg') then
    begin
      jpg := TJPEGImage.Create;
      try
        jpg.LoadFromFile(ImageFileName);
        png := TPngImage.CreateBlank(COLOR_RGBALPHA, 8, jpg.Width, jpg.Height);
        try
          png.Canvas.Draw(0, 0, jpg);
          for i := 0 to png.Height-1 do
            FillChar(png.AlphaScanline[i]^, png.Width, ImageAlpha);
          Image.Assign(png);
        finally
          png.Free;
        end;
      finally
        jpg.Free;
      end;
    end
    else if ext = '.bmp' then
    begin
      bmp := TBitmap.Create;
      try
        bmp.LoadFromFile(ImageFileName);

        png := TPngImage.CreateBlank(COLOR_RGBALPHA, 8, bmp.Width, bmp.Height);
        try
          png.Canvas.Draw(0, 0, bmp);
          for i := 0 to png.Height-1 do
            FillChar(png.AlphaScanline[i]^, png.Width, ImageAlpha);
          Image.Assign(png);
        finally
          png.Free;
        end;
      finally
        bmp.Free;
      end;
    end
    else
      raise Exception.Create('Unsupported image format');
  end;

var
  Image: TPngImage;
  reg: TRegistry;
  buf: TBitmap;
  keys: TStringList;
  i: Integer;
  s: string;
  color: TColor;
begin
  BufferDict.Clear;
  try
    Image := TPngImage.Create;
    try
      LoadImage(Image);

      ImageSize.cx := Image.Width;
      ImageSize.cy := Image.Height;

      reg := TRegistry.Create;
      try
        reg.RootKey := HKEY_CURRENT_USER;
        if not reg.OpenKeyReadOnly(HighlightRegKey) then Exit;

        keys := TStringList.Create;
        try
          reg.GetKeyNames(keys);
          for i := 0 to keys.Count-1 do
          begin
            if not reg.OpenKeyReadOnly(HighlightRegKey + keys[i]) then Continue;

            s := reg.ReadString(HighlightRegName);
            color := StringToColor(s);

            if BufferDict.ContainsKey(color) then Continue;

            buf := TBitmap.Create;
            try
              buf.Width := Image.Width;
              buf.Height := Image.Height;
              buf.Canvas.Brush.Color := color;
              buf.Canvas.FillRect(Rect(0, 0, buf.Width, buf.Height));
              buf.Canvas.Draw(0, 0, Image);
              BufferDict.Add(color, buf);
            except
              buf.Free;
            end;
          end;
        finally
          keys.Free;
        end;
      finally
        reg.Free;
      end;
    finally
      Image.Free;
    end;
  except
    BufferDict.Clear;
    Exit;
  end;

  RefreshEditors;
end;

var
  ItaIDEMenu: TMenuItem;

procedure ItaIDEImageMenuClick(Self: TMenuItem; Sender: TMenuItem);
begin
  ShowConfigDialog;
end;

procedure RegisterMenu;
var
  nta: INTAServices;
  i, j: Integer;
  method: TMethod;
begin
  nta := BorlandIDEServices as INTAServices;
  for i := 0 to nta.MainMenu.Items.Count-1 do
  begin
    if nta.MainMenu.Items[i].Name <> 'ToolsMenu' then Continue;

    for j := 0 to nta.MainMenu.Items[i].Count-1 do
    begin
      if nta.MainMenu.Items[i][j].Name <> 'ToolsToolsItem' then Continue;

      ItaIDEMenu := TMenuItem.Create(nil);
      ItaIDEMenu.AutoHotkeys := maManual;
      ItaIDEMenu.Caption := 'IDEݒ';
      method.Code := @ItaIDEImageMenuClick;
      method.Data := ItaIDEMenu;
      ItaIDEMenu.OnClick := TNotifyEvent(method);
      nta.MainMenu.Items[i].Insert(j, ItaIDEMenu);

      Break;
    end;

    Break;
  end;
end;

procedure UnregisterMenu;
begin
  ItaIDEMenu.Free;
end;

initialization
  BufferDict := TObjectDictionary<TColor,TBitmap>.Create([doOwnsValues]);
  LoadSettings;
  LoadBackgroundImage;
  if not Hook then Exit;
  RegisterMenu;
  TempBuffer := TBitmap.Create;
finalization
  UnregisterMenu;
  Unhook;
  BufferDict.Free;
  TempBuffer.Free;
end.
